"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const _ = require("lodash");
const util_1 = require("util");
function visitType(type, context) {
    const flagLiteral = getLiteralFromFlags(type);
    if (flagLiteral)
        return flagLiteral;
    const symbolLiteral = getLiteralFromSymbol(type);
    if (symbolLiteral)
        return symbolLiteral;
    if (type.isUnion()) {
        const compoundTypes = type.types;
        return visitType(compoundTypes[0], context);
    }
    if (type.isIntersection()) {
        const typeChecker = context.checker;
        const compoundTypes = type.types;
        const allProperties = _.flatten(compoundTypes
            .map(type => typeChecker.getPropertiesOfType(type)
            .map(property => getPropertyDefaultValue(property, context))
            .filter(value => !util_1.isUndefined(value))));
        return ts.createObjectLiteral(allProperties);
    }
    const typeChecker = context.checker;
    const properties = typeChecker.getPropertiesOfType(type);
    const objectDefaultValues = properties
        .map(property => getPropertyDefaultValue(property, context))
        .filter(value => !util_1.isUndefined(value));
    return ts.createObjectLiteral(objectDefaultValues);
}
exports.visitType = visitType;
function getLiteralFromSymbol(type) {
    const symbol = type.getSymbol();
    if (!symbol)
        return undefined;
    switch (symbol.name) {
        case 'Date':
            return ts.createNew(ts.createIdentifier("Date"), undefined, undefined);
        default:
            return undefined;
    }
}
const flagToLiteralMap = {
    [ts.TypeFlags.String]: () => ts.createLiteral(''),
    [ts.TypeFlags.Number]: () => ts.createLiteral(0),
    [ts.TypeFlags.Boolean]: () => ts.createLiteral(false),
    [ts.TypeFlags.StringLiteral]: (type) => ts.createLiteral(type.value),
    [ts.TypeFlags.NumberLiteral]: (type) => ts.createLiteral(type.value),
    [ts.TypeFlags.BooleanLiteral]: () => ts.createLiteral(false),
    [ts.TypeFlags.Void]: () => ts.createNull(),
    [ts.TypeFlags.Null]: () => ts.createNull(),
    [ts.TypeFlags.Object]: (type) => {
        return type.getProperty('length') ? ts.createArrayLiteral([]) : undefined;
    }
};
function getLiteralFromFlags(type) {
    const flags = type.getFlags();
    const literalConverter = flagToLiteralMap[flags];
    return literalConverter ? literalConverter(type) : undefined;
}
function getPropertyDefaultValue(property, context) {
    const propertyNode = property.valueDeclaration;
    const propertyNodeChildren = propertyNode.getChildren();
    let literal = undefined;
    propertyNodeChildren.forEach((child, _ind) => {
        const typeChecker = context.checker;
        if (ts.isTypeReferenceNode(child)) {
            if (property.name === '__@iterator') {
                literal = ts.createObjectLiteral();
                return;
            }
            const type = typeChecker.getTypeFromTypeNode(child);
            literal = visitType(type, context);
            return;
        }
        switch (child.kind) {
            case ts.SyntaxKind.StringKeyword:
                literal = ts.createLiteral('');
                break;
            case ts.SyntaxKind.BooleanKeyword:
                literal = ts.createLiteral(false);
                break;
            case ts.SyntaxKind.NumberKeyword:
                literal = ts.createLiteral(0);
                break;
            case ts.SyntaxKind.NullKeyword:
                literal = ts.createNull();
                break;
            case ts.SyntaxKind.UndefinedKeyword:
                literal = ts.createNull();
                break;
            case ts.SyntaxKind.ArrayType:
                literal = ts.createArrayLiteral([]);
                break;
            case ts.SyntaxKind.Identifier:
                if (ts.isIdentifier(child)) {
                }
                break;
        }
    });
    return literal ? ts.createPropertyAssignment(property.name, literal) : undefined;
}
//# sourceMappingURL=visitor-generate.js.map