"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const tsutils = require("tsutils");
function createPropertyCheck(accessor, property, type, optional, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        const propertyAccessor = ts.createElementAccess(accessor, property);
        const expression = visitType(type, propertyAccessor, Object.assign({}, visitorContext, { mode: { type: 'type-check' } }));
        if (optional) {
            return ts.createBinary(ts.createLogicalNot(ts.createBinary(property, ts.SyntaxKind.InKeyword, accessor)), ts.SyntaxKind.BarBarToken, expression);
        }
        else {
            return ts.createBinary(ts.createBinary(property, ts.SyntaxKind.InKeyword, accessor), ts.SyntaxKind.AmpersandAmpersandToken, expression);
        }
    }
    else {
        return visitType(type, accessor, Object.assign({}, visitorContext, { mode: { type: 'type-check' } }));
    }
}
function visitPropertyName(node, _accessor, _visitorContext) {
    if (ts.isIdentifier(node)) {
        return ts.createStringLiteral(node.text);
    }
    else if (ts.isStringLiteral(node)) {
        return ts.createStringLiteral(node.text);
    }
    else if (ts.isNumericLiteral(node)) {
        return ts.createStringLiteral(node.text);
    }
    else {
        return node.expression;
    }
}
function visitPropertySignature(node, accessor, visitorContext) {
    if (node.type === undefined) {
        throw new Error('Visiting property without type.');
    }
    const type = visitorContext.checker.getTypeFromTypeNode(node.type);
    return createPropertyCheck(accessor, visitPropertyName(node.name, accessor, visitorContext), type, node.questionToken !== undefined, visitorContext);
}
function visitDeclaration(node, accessor, visitorContext) {
    if (ts.isPropertySignature(node)) {
        return visitPropertySignature(node, accessor, visitorContext);
    }
    else if ((node.kind & ts.SyntaxKind.MethodSignature) !== 0) {
        throw new Error('Encountered a method declaration, but methods are not supported. Please check the README.');
    }
    else {
        throw new Error('Unsupported declaration kind: ' + node.kind);
    }
}
function visitTupleObjectType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        if (type.typeArguments === undefined) {
            throw new Error('Expected tuple type to have type arguments.');
        }
        const conditions = [
            ts.createStrictEquality(ts.createPropertyAccess(accessor, ts.createIdentifier('length')), ts.createNumericLiteral(type.typeArguments.length.toString()))
        ];
        for (let i = 0; i < type.typeArguments.length; i++) {
            conditions.push(visitType(type.typeArguments[i], ts.createElementAccess(accessor, i), visitorContext));
        }
        return conditions.reduce((condition, expression) => ts.createBinary(condition, ts.SyntaxKind.AmpersandAmpersandToken, expression));
    }
    else {
        throw new Error('visitTupleObjectType should only be called during type-check mode.');
    }
}
function visitArrayObjectType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        const numberIndexType = visitorContext.checker.getIndexTypeOfType(type, ts.IndexKind.Number);
        if (numberIndexType === undefined) {
            throw new Error('Expected array ObjectType to have a number index type.');
        }
        const itemIdentifier = ts.createIdentifier('item');
        return ts.createBinary(ts.createCall(ts.createPropertyAccess(ts.createIdentifier('Array'), ts.createIdentifier('isArray')), undefined, [accessor]), ts.SyntaxKind.AmpersandAmpersandToken, ts.createCall(ts.createPropertyAccess(accessor, ts.createIdentifier('every')), undefined, [
            ts.createArrowFunction(undefined, undefined, [
                ts.createParameter(undefined, undefined, undefined, itemIdentifier)
            ], undefined, undefined, ts.createBlock([
                ts.createReturn(visitType(numberIndexType, itemIdentifier, visitorContext))
            ]))
        ]));
    }
    else {
        throw new Error('visitArrayObjectType should only be called during type-check mode.');
    }
}
function visitPropertySymbol(property, accessor, visitorContext) {
    if ('valueDeclaration' in property) {
        return visitDeclaration(property.valueDeclaration, accessor, visitorContext);
    }
    else {
        const propertyType = property.type;
        const propertyName = property.name;
        const optional = (property.flags & ts.SymbolFlags.Optional) !== 0;
        if (propertyType !== undefined && propertyName !== undefined) {
            return createPropertyCheck(accessor, ts.createStringLiteral(propertyName), propertyType, optional, visitorContext);
        }
        else {
            throw new Error('Expected a valueDeclaration or a property name and type.');
        }
    }
}
function visitRegularObjectType(type, accessor, visitorContext) {
    const properties = visitorContext.checker.getPropertiesOfType(type);
    if (visitorContext.mode.type === 'type-check') {
        const conditions = [];
        conditions.push(ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('object')), ts.createStrictInequality(accessor, ts.createNull()), ts.createLogicalNot(ts.createCall(ts.createPropertyAccess(ts.createIdentifier('Array'), ts.createIdentifier('isArray')), undefined, [accessor])));
        for (const property of properties) {
            conditions.push(visitPropertySymbol(property, accessor, visitorContext));
        }
        const stringIndexType = visitorContext.checker.getIndexTypeOfType(type, ts.IndexKind.String);
        if (stringIndexType) {
            const keyIdentifier = ts.createIdentifier('key');
            const itemAccessor = ts.createElementAccess(accessor, keyIdentifier);
            conditions.push(ts.createCall(ts.createPropertyAccess(ts.createCall(ts.createPropertyAccess(ts.createIdentifier('Object'), ts.createIdentifier('keys')), undefined, [accessor]), ts.createIdentifier('every')), undefined, [
                ts.createArrowFunction(undefined, undefined, [
                    ts.createParameter(undefined, undefined, undefined, keyIdentifier)
                ], undefined, undefined, ts.createBlock([
                    ts.createReturn(visitType(stringIndexType, itemAccessor, visitorContext))
                ]))
            ]));
        }
        return conditions.reduce((condition, expression) => ts.createBinary(condition, ts.SyntaxKind.AmpersandAmpersandToken, expression));
    }
    else if (visitorContext.mode.type === 'keyof') {
        return properties
            .map((property) => ts.createStrictEquality(accessor, ts.createStringLiteral(property.name)))
            .reduce((condition, expression) => ts.createBinary(condition, ts.SyntaxKind.BarBarToken, expression), ts.createFalse());
    }
    else {
        const indexType = visitorContext.mode.indexType;
        return properties
            .map((property) => {
            const propertyNameAccessor = ts.createStringLiteral(property.name);
            const propertyNameExpression = visitType(indexType, propertyNameAccessor, Object.assign({}, visitorContext, { mode: { type: 'type-check' } }));
            const propertyTypeExpression = visitPropertySymbol(property, accessor, visitorContext);
            return ts.createBinary(ts.createLogicalNot(propertyNameExpression), ts.SyntaxKind.BarBarToken, propertyTypeExpression);
        })
            .reduce((condition, expression) => ts.createBinary(condition, ts.SyntaxKind.BarBarToken, expression), ts.createFalse());
    }
}
function visitObjectType(type, accessor, visitorContext) {
    const mappers = [];
    (function checkBaseTypes(type) {
        if (tsutils.isTypeReference(type) && tsutils.isInterfaceType(type.target)) {
            const baseTypes = visitorContext.checker.getBaseTypes(type.target);
            for (const baseType of baseTypes) {
                if (tsutils.isTypeReference(baseType) && baseType.target.typeParameters !== undefined && baseType.typeArguments !== undefined) {
                    const typeParameters = baseType.target.typeParameters;
                    const typeArguments = baseType.typeArguments;
                    mappers.push((source) => {
                        for (let i = 0; i < typeParameters.length; i++) {
                            if (source === typeParameters[i]) {
                                return typeArguments[i];
                            }
                        }
                    });
                    checkBaseTypes(baseType);
                }
            }
        }
    })(type);
    let targetType;
    if (tsutils.isTypeReference(type)) {
        if (type.target.typeParameters !== undefined && type.typeArguments !== undefined) {
            const typeParameters = type.target.typeParameters;
            const typeArguments = type.typeArguments;
            mappers.push((source) => {
                for (let i = 0; i < typeParameters.length; i++) {
                    if (source === typeParameters[i]) {
                        return typeArguments[i];
                    }
                }
            });
        }
        targetType = type.target;
    }
    else {
        targetType = type;
    }
    const mapper = mappers.reduce((previous, next) => (source) => previous(source) || next(source), () => undefined);
    let expression;
    visitorContext.typeMapperStack.push(mapper);
    if (tsutils.isTupleType(targetType)) {
        expression = visitTupleObjectType(targetType, accessor, visitorContext);
    }
    else if (visitorContext.checker.getIndexTypeOfType(targetType, ts.IndexKind.Number)) {
        expression = visitArrayObjectType(targetType, accessor, visitorContext);
    }
    else {
        expression = visitRegularObjectType(targetType, accessor, visitorContext);
    }
    visitorContext.typeMapperStack.pop();
    return expression;
}
function visitLiteralType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        if (typeof type.value === 'string') {
            return ts.createStrictEquality(accessor, ts.createStringLiteral(type.value));
        }
        else if (typeof type.value === 'number') {
            return ts.createStrictEquality(accessor, ts.createNumericLiteral(type.value.toString()));
        }
        else {
            throw new Error('Type value is expected to be a string or number.');
        }
    }
    else {
        throw new Error('visitLiteralType should only be called during type-check mode.');
    }
}
function visitUnionOrIntersectionType(type, accessor, visitorContext) {
    let token;
    if (tsutils.isUnionType(type)) {
        if (visitorContext.mode.type === 'keyof' || visitorContext.mode.type === 'indexed-access') {
            token = ts.SyntaxKind.AmpersandAmpersandToken;
        }
        else {
            token = ts.SyntaxKind.BarBarToken;
        }
    }
    else if (tsutils.isIntersectionType(type)) {
        if (visitorContext.mode.type === 'keyof' || visitorContext.mode.type === 'indexed-access') {
            token = ts.SyntaxKind.BarBarToken;
        }
        else {
            token = ts.SyntaxKind.AmpersandAmpersandToken;
        }
    }
    else {
        throw new Error('UnionOrIntersection type is expected to be a Union or Intersection type.');
    }
    return type.types
        .map((type) => visitType(type, accessor, visitorContext))
        .reduce((condition, expression) => ts.createBinary(condition, token, expression));
}
function visitBooleanLiteral(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(accessor, type.intrinsicName === 'true'
            ? ts.createTrue()
            : ts.createFalse());
    }
    else {
        throw new Error('visitBooleanLiteral should only be called during type-check mode.');
    }
}
function visitNonPrimitiveType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        const intrinsicName = type.intrinsicName;
        let conditions;
        if (intrinsicName === 'object') {
            conditions = [
                ts.createStrictInequality(ts.createTypeOf(accessor), ts.createStringLiteral('boolean')),
                ts.createStrictInequality(ts.createTypeOf(accessor), ts.createStringLiteral('number')),
                ts.createStrictInequality(ts.createTypeOf(accessor), ts.createStringLiteral('string'))
            ];
        }
        else {
            throw new Error(`Unsupported non-primitive with intrinsic name: ${intrinsicName}.`);
        }
        return conditions.reduce((condition, expression) => ts.createBinary(condition, ts.SyntaxKind.AmpersandAmpersandToken, expression));
    }
    else {
        throw new Error('visitNonPrimitiveType should only be called during type-check mode.');
    }
}
function visitTypeParameter(type, accessor, visitorContext) {
    const typeMapper = visitorContext.typeMapperStack.reduceRight((previous, next) => (source) => previous(source) || next(source), () => undefined);
    if (typeMapper === undefined) {
        throw new Error('Unbound type parameter, missing type mapper.');
    }
    const mappedType = typeMapper(type) || type.getDefault();
    if (mappedType === undefined) {
        throw new Error('Unbound type parameter, missing type node.');
    }
    return visitType(mappedType, accessor, visitorContext);
}
function visitIndexType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        const indexedType = type.type;
        if (indexedType === undefined) {
            throw new Error('Could not get indexed type of index type.');
        }
        return visitType(indexedType, accessor, Object.assign({}, visitorContext, { mode: { type: 'keyof' } }));
    }
    else {
        throw new Error('visitIndexType should only be called during type-check mode.');
    }
}
function visitIndexedAccessType(type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return visitType(type.objectType, accessor, Object.assign({}, visitorContext, { mode: { type: 'indexed-access', indexType: type.indexType } }));
    }
    else {
        throw new Error('visitIndexedAccessType should only be called during type-check mode.');
    }
}
function visitAny(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createTrue();
    }
    else if (visitorContext.mode.type === 'keyof') {
        return ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('string'));
    }
    else {
        return ts.createTrue();
    }
}
function visitUnknown(_type, _accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createTrue();
    }
    else if (visitorContext.mode.type === 'keyof') {
        return ts.createFalse();
    }
    else {
        throw new Error('visitUnknown should only be called during type-check or keyof mode.');
    }
}
function visitNever(_type, _accessor, _visitorContext) {
    return ts.createFalse();
}
function visitNull(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(accessor, ts.createNull());
    }
    else {
        throw new Error('visitNull should only be called during type-check mode.');
    }
}
function visitUndefined(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(accessor, ts.createIdentifier('undefined'));
    }
    else {
        throw new Error('visitUndefined should only be called during type-check mode.');
    }
}
function visitNumber(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('number'));
    }
    else {
        throw new Error('visitNumber should only be called during type-check mode.');
    }
}
function visitBigInt(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('bigint'));
    }
    else {
        throw new Error('visitBigInt should only be called during type-check mode.');
    }
}
function visitBoolean(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('boolean'));
    }
    else {
        throw new Error('visitBoolean should only be called during type-check mode.');
    }
}
function visitString(_type, accessor, visitorContext) {
    if (visitorContext.mode.type === 'type-check') {
        return ts.createStrictEquality(ts.createTypeOf(accessor), ts.createStringLiteral('string'));
    }
    else {
        throw new Error('visitString should only be called during type-check mode.');
    }
}
function visitType(type, accessor, visitorContext) {
    if ((ts.TypeFlags.Any & type.flags) !== 0) {
        return visitAny(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Unknown & type.flags) !== 0) {
        return visitUnknown(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Never & type.flags) !== 0) {
        return visitNever(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Null & type.flags) !== 0) {
        return visitNull(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Undefined & type.flags) !== 0) {
        return visitUndefined(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Number & type.flags) !== 0) {
        return visitNumber(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.BigInt & type.flags) !== 0) {
        return visitBigInt(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Boolean & type.flags) !== 0) {
        return visitBoolean(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.String & type.flags) !== 0) {
        return visitString(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.BooleanLiteral & type.flags) !== 0) {
        return visitBooleanLiteral(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.TypeParameter & type.flags) !== 0) {
        return visitTypeParameter(type, accessor, visitorContext);
    }
    else if (tsutils.isObjectType(type)) {
        if ((ts.ObjectFlags.Class & type.objectFlags) !== 0) {
            throw new Error('Classes cannot be validated. Please check the README.');
        }
        else {
            return visitObjectType(type, accessor, visitorContext);
        }
    }
    else if (tsutils.isLiteralType(type)) {
        return visitLiteralType(type, accessor, visitorContext);
    }
    else if (tsutils.isUnionOrIntersectionType(type)) {
        return visitUnionOrIntersectionType(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.NonPrimitive & type.flags) !== 0) {
        return visitNonPrimitiveType(type, accessor, visitorContext);
    }
    else if ((ts.TypeFlags.Index & type.flags) !== 0) {
        return visitIndexType(type, accessor, visitorContext);
    }
    else if (tsutils.isIndexedAccessType(type)) {
        return visitIndexedAccessType(type, accessor, visitorContext);
    }
    else {
        throw new Error('Could not generate type-check; unsupported type with flags: ' + type.flags);
    }
}
exports.visitType = visitType;
function visitUndefinedOrType(type, accessor, visitorContext) {
    return ts.createBinary(ts.createStrictEquality(accessor, ts.createIdentifier('undefined')), ts.SyntaxKind.BarBarToken, visitType(type, accessor, visitorContext));
}
exports.visitUndefinedOrType = visitUndefinedOrType;
//# sourceMappingURL=visitor.js.map